// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Ccf.h	--- interface between the CAL and the CCF
//
//	DESCRIPTION:
//		This defines the interface between the CCF and the CAL
//
//	CHANGE HISTORY:
//		UL, 1997-10-10: 1.0.0: - skeleton
//		UL, 1999-03-09: 1.1.0: - some methods in CifEntry now return
//								 constant references to data members instead
//								 of references
//		UL, 1999-08-10: 1.2.0: - added Ccf::cifName()
//		UL, 2000-10-05: 2.0.0: - added Ccf::cifDirName() / Ccf::ccfSearchPath()
//							   - added CifEntry::open()
//		UL, 2000-12-18: 2.1.0: - added CcfConstituentBlock
//		UL, 2003-05-23: 2.2.0: - added Ccf::SAS_CCFFILES
//		RDS, 2006-02-07: 2.4.0: - added EFFAREACORR
//		RDS, 2006-03-31: 2.5.0: - added COOLPIX
//		RDS, 2009-04-12: 2.6.0: - added SCREENTHRESH, SPECQUAL
//	AUTHOR:
//		U. Lammers, Fri Oct 10 12:43:51 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/Ccf.h,v 1.17 2013/08/22 15:10:59 rsaxton Exp $
//

#ifndef CCF_H
#define CCF_H

#include <string>
#include <set>
#include <utility>

#include "Dal.h"
#include "Xmm.h"
#include "STime.h"
#include "StringIntMap.h"
#include "CalChild.h"

class CifEntry;	// forward declaration

class Ccf {
public:
	Ccf(const std::string cifFileName = "");
		// open a CIF defining the CCF

	//
	//	whenever needs to be added here, make sure you add it to the end
	enum CalDataTypes { NONE = -1,
		AbsCoefs = 0,
		ADUConv,
		AstroMet,
		Background,
		BadPix,
		BoreSight,
		CCDrmf,
		CTI,
		CalIndex,
		CalSourceData,
		ClockPatterns,
		ColorTrans,
		Contamination,
		CrossPSF,
		DarkCurrent,
		DarkFrame,
		DiffuseGala,
		Efficiency,
		EvtSizeDist,
		EXAFS,
		FilterTransV,
		FilterTransX,
		HKParmInt,
		LeapSeconds,
		LightLeak,
		LinCoord,
		LineSpreadFunc,
		LargeScaleSens,
		MiscData,
		ModeParam,
		PatternLib,
		PixToPixSens,
		PhotToNat,
		PhotVsLam,
		PixToLam,
		PSF1DRB,
		PSFPileUp,
		QuantumEf,
		QuickMag,
		Redist,
		RedistPileUp,
		StrayLight,
		TimeCorr,
		TimeJumpTol,
		VAreaEf,
		VPSF,
		VSLAreaEf,
		VSLDist,
		XAreaEf,
		XEncirEn,
		XPSF,
		XSLAreaEf,
		XSLDist,
		Zodiacal,
		DiagnosticOffset,
		GratingData,
		GrismData,
        Reject,
        TemplateBckgnd,
		EffAreaCorr,
        CoolPix,
        ScreenThresh,
        SpecQual,
        Ticlosedodi,
        Buclosedodi,
        SAACorr,
        FWC,
        FWC_EFF,
        FLARE,
        SWCX,
        SCALEFACTORS,
        SPDETMAP,
        SPATIALCTI
	};

	//  DO NOT FORGET to update the following number of the numbers
	//  of elements in the above enum has changed!
	static const unsigned N_CALDATATYPES = SPDETMAP+1;

	#define TYPE_STR \
		"ABSCOEFS", "ADUCONV", "ASTROMET", \
        "BACKGROUND", "BADPIX", \
		"BORESIGHT", "CCDRMF", "CTI", "CALINDEX", "CALSOURCEDATA", \
		"CLOCKPATTERNS", \
		"COLORTRANS", "CONTAMINATION", "CROSSPSF", "DARKCURRENT", "DARKFRAME", \
		"DIFFUSEGALA", "EFFICIENCY", "EVTSIZEDIST", "EXAFS", "FILTERTRANSV", \
		"FILTERTRANSX", "HKPARMINT", "LEAPSECONDS", "LIGHTLEAK", "LINCOORD", \
		"LINESPREADFUNC", "LARGESCALESENS", "MISCDATA", "MODEPARAM", \
		"PATTERNLIB", "PIXTOPIXSENS", "PHOTTONAT", "PHOTVSLAM", "PIXTOLAM", \
		"PSF1DRB", "PSFPILEUP", "QUANTUMEF", "QUICKMAG", "REDIST", \
		"REDISTPILEUP", "STRAYLIGHT", "TIMECORR", "TIMEJUMPTOL", \
                "VAREAEF", "VPSF", \
		"VSLAREAEF", "VSLDIST", "XAREAEF", "XENCIREN", "XPSF", "XSLAREAEF", \
		"XSLDIST", "ZODIACAL", "DIAGNOSTIC_OFFSETS", "GRATING_DATA", \
		"GRISMCAL", "REJECT","TEMPLATEBCKGND","EFFAREACORR","COOLPIX", \
                "SCREENTHRESH", "SPECQUAL", "TICLOSEDODI", "BUCLOSEDODI", \
                "SAACORR", "FWC", "FWC_EFF", "FLARE", "SWCX", "SCALEFACTORS", \
                "SPDETMAP", "SPATIALCTI"

	//
	//	Ccf::toTypeId("ZODIACAL") == Ccf::ZODICAL
	//	Ccf::toString(Ccf::ZODIACAL) == "ZODIACAL"
	//
	static CalDataTypes toTypeId(const std::string &typeString);
	static const std::string &toString(CalDataTypes dataTypeId);
	static const std::string &toString(unsigned dataTypeId);


	static const char *const SAS_CCF;				//	= "SAS_CCF";
		//	environment variable pointing to block "CALINDEX" containing
		//	the CIF; can also be the name of a directory in which case the
		//	CIF is supposed to have a default name SAS_CIFDFNAME

	static const char *const SAS_DEFAULT_CIFNAME;	//	= "ccf.cif"
		//	default name of the CIF

	static const char *const SAS_CCFPATH;			//	= "SAS_CCFPATH"
		//	environment variable containing colon separated list of
		//	directories; this defines the search path for locating
		//	CCF constituents

	static const char *const SAS_CCFFILES;			//	= "SAS_CCFFILES"
		//	environment variable containing white-space separated list
		//	of CCF constituent replacements; if given without path
		//	name the file is searched along SAS_CCFPATH

	//
	//	name of CIF / directory name of CIF / CCF constituent search path
	//
	const std::string &cifName() const { return _cifFileName; }
	const std::string &cifDirName() const { return _cifDirName; }
	const std::string &ccfSearchPath() const { return _ccfSearchPath; }

	//	set the search path for CCF constituents; must be a colon
	//	separated list of directory names
	void ccfSearchPath(const std::string &path) { _ccfSearchPath = path; }

	//
	//	register new CifEntry
	//
	virtual void registerNewCifEntry(const std::string &ccfSetName) = 0;

	//	query for a calibration file of a particular type corresponding to an
	//	instrument
	virtual bool hasEntry(const XmmInstrument *, Ccf::CalDataTypes) const =0;
	virtual CifEntry *entry(const XmmInstrument *, Ccf::CalDataTypes,
							bool generateError = true) const = 0;

	virtual void print(std::ostream &) const;
	friend std::ostream &operator<<(std::ostream &, const Ccf *);

	virtual ~Ccf();						// destructor

	typedef pair<Xmm::Instrument, Ccf::CalDataTypes> CcfConstituent;
	typedef std::map<CcfConstituent, CifEntry *, std::less<CcfConstituent> >
			Cif;

protected:
	std::string _cifFileName,				//	name of current CIF (full path)
			    _cifDirName,				//	directory
			    _ccfSearchPath;				//	search path for constituents
			
private:
	static StringIntMap<Ccf::CalDataTypes>	_calDataTypes;

	friend class CalviewGui;
	friend class ShowCcf;
	virtual const Cif &cif() const = 0;
}; // Ccf

//
//	print CCF to a stream
//
std::ostream &operator<<(std::ostream &, const Ccf *);


template<class T>
class CcfConstituentBlock {
public:
	CcfConstituentBlock(const DataSet *set, const std::string &name,
						unsigned acceptAlgoIdsLessEqThan = 0,
						bool treatBlockAsOptional = false);
	CcfConstituentBlock(const DataSet *set, unsigned nr,
						unsigned acceptAlgoIdsLessEqThan = 0,
						bool treatBlockAsOptional = false);
		//
		//	PURPOSE:
		//		Construct object; a CcfConstituentBlock is a block
		//		(table or array) in a CCF constituent
		//
		//	ARGUMENTS:
		//		set		: data set containing the block
		//		name	: name of the block to access
		//		number	: number of block to access
		//		acceptAlgoIdsLessEqThan:
		//				  each block is associated with a dedicated calibration
		//				  algorithm identified through its ID (>=0);
		//				  the ID is encoded in attribute
		//					ALGOID
		//				  in the table; a missing attribute is taken as
		//				  ALGOID=0; all IDs less than or equal to
		//				  <acceptAlgoIdsLessEqThan> will be accepted otherwise
		//				  an error condition is raised
		//		treatBlockAsOptional:
		//				  if set to true the named block can be missing from
		//				  the data set; if set to false (d/f) access to a
		//				  missing block will cause an error condition
		//


	const T *block() const { return _block; }
		//
		//	PURPOSE:
		//		Access block associated with this object
		//


	unsigned algoid() const { return _algoid; }
		//
		//	PURPOSE:
		//		Return ALGOID of the block
		//


	~CcfConstituentBlock() {}
		//
		//	PURPOSE:
		//		Destroy object
		//

private:
	unsigned	_algoid;
	const T		*_block;
};


typedef CcfConstituentBlock<Table>	CcfConstituentTable;
typedef CcfConstituentBlock<Array>	CcfConstituentArray;

class CifEntry : public CalChild<Ccf> {
public:
	//	construct from data tuple
	CifEntry(Xmm::Instrument instrument, Ccf::CalDataTypes calDataType,
			 int issue, const std::string &fileName,
			 const std::string &valDate, const std::string &subDate,
			 const std::string &extKey, const std::string &extSeq,
			 const Ccf *ccf = 0)
			: CalChild<Ccf>(ccf),
			 _instrument(instrument),
			 _calDataType(calDataType),
			 _issue(issue),
			 _fileName(fileName),
			 _validityDate(valDate),
			 _submissionDate(subDate),
			 _extensionKey(extKey),
			 _extensionSequence(extSeq)
		   {}

	//	construct from calibration file name and CCF
	CifEntry(const std::string &calFileName, const Ccf *ccf);

	//	copy constructor
	CifEntry &operator=(const CifEntry &entry);
	CifEntry(const CifEntry &entry);

	//	query block number/name
	bool hasBlock(const std::string &truncatedBlockName) const;
	std::string queryBlockName(const std::string &truncatedBlockName,
						  unsigned *blockNr = 0) const;

	Xmm::Instrument		instrument() const { return _instrument; }
	Ccf::CalDataTypes	calDataType() const { return _calDataType; }
	int					issue() const { return _issue; }
	std::string			fileName() const { return _fileName; }
	const STime			&validityDate() const { return _validityDate; }
	const STime			&submissionDate() const { return _submissionDate;}
	const std::string	&extensionKey() const { return _extensionKey; }
	const std::string	&extensionSequence() const {
												return _extensionSequence; }

	DataSet *open(DataSetServer::MemoryModel memModel = DataSetServer::Low,
				  bool logAccess = true) const;
		//
		//	PURPOSE:
		//		Try to open the CCF constituent in Read-Mode using the
		//		file name specified in the CIF. If this fails and SAS_CCFPATH
		//		is defined try to locate the file along the path(s); if
		//		SAS_CCFPATH	is undefined raise an error condition
		//
		//	ARGUMENTS:
		//		memModel	: DAL memory model to open the data set with
		//		logAccess	: log this access in global CcfAccessLog?
		//

	//	destructor
	~CifEntry();

	friend std::ostream &operator<<(std::ostream &, const CifEntry &);

private:
	//	default constructor
	explicit CifEntry() : CalChild<Ccf>(0) {}

	// data members
	Xmm::Instrument		_instrument;			// instrument name
	Ccf::CalDataTypes	_calDataType;			// file type
	int					_issue;					// file issue number
	mutable std::string	_fileName;				// name on local file system
	STime				_validityDate,			// date of validity
						_submissionDate;		// date of submission
	std::string			_extensionKey,			// key of type of extensions
						_extensionSequence;		// truncated extension names
};

std::ostream &operator<<(std::ostream &, const CifEntry &);

//
//	CIF : a Cif is a set of CifEntries
//
class Cif : public std::set<CifEntry> {};

//
//	CcfAccessLog: class used to log CCF accesses
//
class CcfAccessLog : public std::set<CifEntry> {
public:
	CcfAccessLog();
	void reset();
	std::pair<std::set<CifEntry>::iterator, bool> insert(const CifEntry &);
	// check CifEntry vs list of expired CCF constituents
	void evaluate(const CifEntry &) const;
	void evaluate() const;
	~CcfAccessLog();
	static CcfAccessLog &instance();

private:
	static CcfAccessLog	_instance;

	std::map<std::string, std::string> invalidCcf_;

};

std::ostream &operator<<(std::ostream &, const CcfAccessLog &);

#endif
