// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	StringPool.h --- a string<vector> with additional access functionality
//
//	DESCRIPTION:
//		StringPool can be used synonymous to vector<string>; in addition
//		it offers a method
//			string expand(const &keyString)
//		which realizes member access via key strings; a keystring must
//		be a unique abbreviation of the sought string; see below for
//		more explanations.
//
//	USAGE EXAMPLES:
//		see below
//
//	CHANGE HISTORY:
//		UL, 1998-06-02: 1.0.0: - first version
//		UL, 1999-08-05: 1.1.0: - removed os/is arguments from constructors;
//							     input stream is now always cin and output
//								 stream errstr from error packages
//
//	AUTHOR:
//		U. Lammers, Tue Jun  2 16:06:12 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/StringPool.h,v 1.7 2009/06/23 13:46:02 aibarra Exp $
///

#ifndef STRING_POOL_H
#define STRING_POOL_H

#include <string>
#include <vector>

typedef class NodeData *NodeP;

class StringPool : public std::vector<std::string> {
public:
	typedef enum { SILENT, WARN, ABORT, CORRECT } Verbosity;
		//	error behavior of expand-method; see below for explanation

	StringPool(Verbosity verbosity = CORRECT);
		//	construct empty string pool

	StringPool(const char *stringVec[],
			   Verbosity verbosity = CORRECT);
		//	fill the pool from a vector of char *'s; end must be marked by
		//	null pointer

	StringPool(const std::vector<std::string> &stringVec,
			   Verbosity verbosity = CORRECT);
		//	fill the pool from a proper string vector

	std::string expand(const std::string &keyString) const;
		//	find a string in the pool via a key string, e.g.
		//		char *v[] = { "aab", "aabc", "abcdef", "xqwert" }
		//		StringPool pool(v);
		//		expand("aab") -> "aab"
		//		expand("ab"), expand("abc"), expand("abcd"), etc. -> "abcdef"
		//		expand("x"), expand("xq"), etc. -> "xqwert"
		//
		//	value of verbosity argument in constructor defines behavior
		//	in the two possible error cases:
		//		1) no-match:
		//				SILENT : return empty string; no error message
		//				WARN   : return empty string; print warning message
		//				ABORT  : WARN-behavior + exit
		//				CORRECT: WARN-behavior + prompt user for new keystring
		//
		//		2) ambigious match:
		//				SILENT : return empty string; no error message
		//				WARN   : print list of possible completions
		//						 and return empty string
		//				ABORT  : WARN-behavior + exit
		//				CORRECT: WARN-behavior + prompt user to give more
		//						 characters for making the exansion unique
		//
		//	output messages/additional input are printed/is read to/from
		//	stream passed to constructor
		//
	const std::string &operator[](unsigned n) const
	{
		const std::vector<std::string> *p = this;
		return (*p)[n];
	}

	const std::string &operator[](const std::string &keyString) const;
		//	same as expand-method except that reference to string instance
		//	in vector is returned

	~StringPool();
		//	destructor

private:
	//	data
	NodeP		_root;
	Verbosity	_verbosity;

private:
	//	methods
	void	init();
	void	addString(const char *s);
	void	completeSingle(const NodeP n, std::string &matchString) const;
	/* AI gcc4 porting
	 * extra qualification 'StringPool::' on member  'completeMulti'
	 */
	std::vector<std::string> completeMulti(const NodeP node) const;
	bool	match(const NodeP n, const char *p, std::string &matchString,
				  NodeP &lastMatched) const;
};

#endif
